@@ -646,6 +646,10 @@ extern int var_max_backoff_time;
 #define DEF_MAX_QUEUE_TIME	"5d"
 extern int var_max_queue_time;
 
+ /*
+  * XXX The default can't be $maximal_queue_lifetime, because that panics
+  * when a non-default maximal_queue_lifetime setting contains no time unit.
+  */
 #define VAR_DSN_QUEUE_TIME	"bounce_queue_lifetime"
 #define DEF_DSN_QUEUE_TIME	"5d"
 extern int var_dsn_queue_time;
@@ -1163,6 +1167,10 @@ extern char *var_smtpd_tls_scache_db;
 #define DEF_SMTPD_TLS_SCACHTIME	"3600s"
 extern int var_smtpd_tls_scache_timeout;
 
+#define VAR_SMTPD_DELAY_OPEN	"smtpd_delay_open_until_valid_rcpt"
+#define DEF_SMTPD_DELAY_OPEN	1
+extern bool var_smtpd_delay_open;
+
 #define VAR_SMTP_TLS_PER_SITE	"smtp_tls_per_site"
 #define DEF_SMTP_TLS_PER_SITE	""
 #define VAR_LMTP_TLS_PER_SITE	"lmtp_tls_per_site"
