@@ -247,7 +247,8 @@ struct QMGR_MESSAGE {
     VSTREAM *fp;			/* open queue file or null */
     int     refcount;			/* queue entries */
     int     single_rcpt;		/* send one rcpt at a time */
-    struct timeval arrival_time;	/* time when queued */
+    struct timeval arrival_time;	/* start of receive transaction */
+    time_t  create_time;		/* queue file create time */
     struct timeval active_time;		/* time of entry into active queue */
     time_t  queued_time;		/* sanitized time when moved to the
 					 * active queue */
