@@ -101,6 +101,7 @@ typedef struct {
     pcre_extra *hints;			/* hints to speed pattern execution */
     char   *replacement;		/* replacement string */
     int     match;			/* positive or negative match */
+    size_t  max_sub;			/* largest $number in replacement */
 } DICT_PCRE_MATCH_RULE;
 
 typedef struct {
@@ -116,6 +117,7 @@ typedef struct {
 typedef struct {
     DICT    dict;			/* generic members */
     DICT_PCRE_RULE *head;
+    VSTRING *expansion_buf;		/* lookup result */
 } DICT_PCRE;
 
 static int dict_pcre_init = 0;		/* flag need to init pcre library */
@@ -124,9 +126,8 @@ static int dict_pcre_init = 0;		/* flag need to init pcre library */
  * Context for $number expansion callback.
  */
 typedef struct {
-    const char *mapname;		/* source dict name */
-    int     lineno;			/* source file line number */
-    VSTRING *expansion_buf;		/* target string buffer */
+    DICT_PCRE *dict_pcre;		/* the dictionary handle */
+    DICT_PCRE_MATCH_RULE *match_rule;	/* the rule we matched */
     const char *lookup_string;		/* string against which we match */
     int     offsets[PCRE_MAX_CAPTURE * 3];	/* Cut substrings */
     int     matches;			/* Count of cuts */
@@ -138,8 +139,8 @@ typedef struct {
 typedef struct {
     const char *mapname;		/* name of regexp map */
     int     lineno;			/* where in file */
-    int     flags;			/* dict_flags */
     size_t  max_sub;			/* Largest $n seen */
+    char   *literal;			/* constant result, $$ -> $ */
 } DICT_PCRE_PRESCAN_CONTEXT;
 
  /*
@@ -162,6 +163,8 @@ typedef struct {
 static int dict_pcre_expand(int type, VSTRING *buf, char *ptr)
 {
     DICT_PCRE_EXPAND_CONTEXT *ctxt = (DICT_PCRE_EXPAND_CONTEXT *) ptr;
+    DICT_PCRE_MATCH_RULE *match_rule = ctxt->match_rule;
+    DICT_PCRE *dict_pcre = ctxt->dict_pcre;
     const char *pp;
     int     n;
     int     ret;
@@ -176,16 +179,16 @@ static int dict_pcre_expand(int type, VSTRING *buf, char *ptr)
 	if (ret < 0) {
 	    if (ret == PCRE_ERROR_NOSUBSTRING)
 		msg_fatal("regexp %s, line %d: replace index out of range",
-			  ctxt->mapname, ctxt->lineno);
+			  dict_pcre->dict.name, match_rule->rule.lineno);
 	    else
 		msg_fatal("regexp %s, line %d: pcre_get_substring error: %d",
-			  ctxt->mapname, ctxt->lineno, ret);
+			dict_pcre->dict.name, match_rule->rule.lineno, ret);
 	}
 	if (*pp == 0) {
 	    myfree((char *) pp);
 	    return (MAC_PARSE_UNDEF);
 	}
-	vstring_strcat(ctxt->expansion_buf, pp);
+	vstring_strcat(dict_pcre->expansion_buf, pp);
 	myfree((char *) pp);
 	return (MAC_PARSE_OK);
     }
@@ -194,7 +197,7 @@ static int dict_pcre_expand(int type, VSTRING *buf, char *ptr)
      * Straight text - duplicate with no substitution.
      */
     else {
-	vstring_strcat(ctxt->expansion_buf, vstring_str(buf));
+	vstring_strcat(dict_pcre->expansion_buf, vstring_str(buf));
 	return (MAC_PARSE_OK);
     }
 }
@@ -204,7 +207,7 @@ static int dict_pcre_expand(int type, VSTRING *buf, char *ptr)
 static void dict_pcre_exec_error(const char *mapname, int lineno, int errval)
 {
     switch (errval) {
-    case 0:
+	case 0:
 	msg_warn("pcre map %s, line %d: too many (...)",
 		 mapname, lineno);
 	return;
@@ -252,7 +255,6 @@ static const char *dict_pcre_lookup(DICT *dict, const char *lookup_string)
     DICT_PCRE_MATCH_RULE *match_rule;
     int     lookup_len = strlen(lookup_string);
     DICT_PCRE_EXPAND_CONTEXT ctxt;
-    static VSTRING *expansion_buf;
     int     nesting = 0;
 
     dict_errno = 0;
@@ -260,6 +262,13 @@ static const char *dict_pcre_lookup(DICT *dict, const char *lookup_string)
     if (msg_verbose)
 	msg_info("dict_pcre_lookup: %s: %s", dict->name, lookup_string);
 
+    /*
+     * Optionally fold the key.
+     */
+    if (dict->fold_buf) {
+	vstring_strcpy(dict->fold_buf, lookup_string);
+	lookup_string = lowercase(vstring_str(dict->fold_buf));
+    }
     for (rule = dict_pcre->head; rule; rule = rule->next) {
 
 	/*
@@ -292,28 +301,31 @@ static const char *dict_pcre_lookup(DICT *dict, const char *lookup_string)
 		continue;			/* pcre_exec failed */
 	    }
 
-	    /* Negative rules can't have any substitutions */
-	    if (!match_rule->match)
+	    /*
+	     * Skip $number substitutions when the replacement text contains
+	     * no $number strings, as learned during the compile time
+	     * pre-scan. The pre-scan already replaced $$ by $.
+	     */
+	    if (match_rule->max_sub == 0)
 		return match_rule->replacement;
 
 	    /*
 	     * We've got a match. Perform substitution on replacement string.
 	     */
-	    if (expansion_buf == 0)
-		expansion_buf = vstring_alloc(10);
-	    VSTRING_RESET(expansion_buf);
-	    ctxt.expansion_buf = expansion_buf;
+	    if (dict_pcre->expansion_buf == 0)
+		dict_pcre->expansion_buf = vstring_alloc(10);
+	    VSTRING_RESET(dict_pcre->expansion_buf);
+	    ctxt.dict_pcre = dict_pcre;
+	    ctxt.match_rule = match_rule;
 	    ctxt.lookup_string = lookup_string;
-	    ctxt.mapname = dict->name;
-	    ctxt.lineno = rule->lineno;
 
 	    if (mac_parse(match_rule->replacement, dict_pcre_expand,
 			  (char *) &ctxt) & MAC_PARSE_ERROR)
 		msg_fatal("pcre map %s, line %d: bad replacement syntax",
 			  dict->name, rule->lineno);
 
-	    VSTRING_TERMINATE(expansion_buf);
-	    return (vstring_str(expansion_buf));
+	    VSTRING_TERMINATE(dict_pcre->expansion_buf);
+	    return (vstring_str(dict_pcre->expansion_buf));
 
 	    /*
 	     * Conditional. XXX We provide space for matched substring info
@@ -392,6 +404,10 @@ static void dict_pcre_close(DICT *dict)
 	}
 	myfree((char *) rule);
     }
+    if (dict_pcre->expansion_buf)
+	vstring_free(dict_pcre->expansion_buf);
+    if (dict->fold_buf)
+	vstring_free(dict->fold_buf);
     dict_free(dict);
 }
 
@@ -493,12 +509,15 @@ static int dict_pcre_prescan(int type, VSTRING *buf, char *context)
     DICT_PCRE_PRESCAN_CONTEXT *ctxt = (DICT_PCRE_PRESCAN_CONTEXT *) context;
     size_t  n;
 
+    /*
+     * Keep a copy of literal text (with $$ already replaced by $) if and
+     * only if the replacement text contains no $number expression. This way
+     * we can avoid having to scan the replacement text at lookup time.
+     */
     if (type == MAC_PARSE_VARNAME) {
-	if (ctxt->flags & DICT_FLAG_NO_REGSUB) {
-	    msg_warn("pcre map %s, line %d: "
-		     "regular expression substitution is not allowed",
-		     ctxt->mapname, ctxt->lineno);
-	    return (MAC_PARSE_ERROR);
+	if (ctxt->literal) {
+	    myfree(ctxt->literal);
+	    ctxt->literal = 0;
 	}
 	if (!alldig(vstring_str(buf))) {
 	    msg_warn("pcre map %s, line %d: non-numeric replacement index \"%s\"",
@@ -513,6 +532,11 @@ static int dict_pcre_prescan(int type, VSTRING *buf, char *context)
 	}
 	if (n > ctxt->max_sub)
 	    ctxt->max_sub = n;
+    } else if (type == MAC_PARSE_LITERAL && ctxt->max_sub == 0) {
+	if (ctxt->literal)
+	    msg_panic("pcre map %s, line %d: multiple literals but no $number",
+		      ctxt->mapname, ctxt->lineno);
+	ctxt->literal = mystrdup(vstring_str(buf));
     }
     return (MAC_PARSE_OK);
 }
@@ -599,14 +623,23 @@ static DICT_PCRE_RULE *dict_pcre_parse_rule(const char *mapname, int lineno,
 	 */
 	prescan_context.mapname = mapname;
 	prescan_context.lineno = lineno;
-	prescan_context.flags = dict_flags;
 	prescan_context.max_sub = 0;
+	prescan_context.literal = 0;
+
+	/*
+	 * The optimizer will eliminate code duplication and/or dead code.
+	 */
+#define CREATE_MATCHOP_ERROR_RETURN(rval) do { \
+	if (prescan_context.literal) \
+	    myfree(prescan_context.literal); \
+	return (rval); \
+    } while (0)
 
 	if (mac_parse(p, dict_pcre_prescan, (char *) &prescan_context)
 	    & MAC_PARSE_ERROR) {
 	    msg_warn("pcre map %s, line %d: bad replacement syntax: "
 		     "skipping this rule", mapname, lineno);
-	    return (0);
+	    CREATE_MATCHOP_ERROR_RETURN(0);
 	}
 
 	/*
@@ -615,14 +648,20 @@ static DICT_PCRE_RULE *dict_pcre_parse_rule(const char *mapname, int lineno,
 	if (prescan_context.max_sub > 0 && regexp.match == 0) {
 	    msg_warn("pcre map %s, line %d: $number found in negative match "
 		   "replacement text: skipping this rule", mapname, lineno);
-	    return (0);
+	    CREATE_MATCHOP_ERROR_RETURN(0);
+	}
+	if (prescan_context.max_sub > 0 && (dict_flags & DICT_FLAG_NO_REGSUB)) {
+	    msg_warn("pcre map %s, line %d: "
+		     "regular expression substitution is not allowed: "
+		     "skipping this rule", mapname, lineno);
+	    CREATE_MATCHOP_ERROR_RETURN(0);
 	}
 
 	/*
 	 * Compile the pattern.
 	 */
 	if (dict_pcre_compile(mapname, lineno, &regexp, &engine) == 0)
-	    return (0);
+	    CREATE_MATCHOP_ERROR_RETURN(0);
 
 	/*
 	 * Save the result.
@@ -631,7 +670,11 @@ static DICT_PCRE_RULE *dict_pcre_parse_rule(const char *mapname, int lineno,
 	    dict_pcre_rule_alloc(DICT_PCRE_OP_MATCH, nesting, lineno,
 				 sizeof(DICT_PCRE_MATCH_RULE));
 	match_rule->match = regexp.match;
-	match_rule->replacement = mystrdup(p);
+	match_rule->max_sub = prescan_context.max_sub;
+	if (prescan_context.literal)
+	    match_rule->replacement = prescan_context.literal;
+	else
+	    match_rule->replacement = mystrdup(p);
 	match_rule->pattern = engine.pattern;
 	match_rule->hints = engine.hints;
 	return ((DICT_PCRE_RULE *) match_rule);
@@ -746,7 +789,10 @@ DICT   *dict_pcre_open(const char *mapname, int unused_flags, int dict_flags)
     dict_pcre->dict.lookup = dict_pcre_lookup;
     dict_pcre->dict.close = dict_pcre_close;
     dict_pcre->dict.flags = dict_flags | DICT_FLAG_PATTERN;
+    if (dict_flags & DICT_FLAG_FOLD_MUL)
+	dict_pcre->dict.fold_buf = vstring_alloc(10);
     dict_pcre->head = 0;
+    dict_pcre->expansion_buf = 0;
 
     if (dict_pcre_init == 0) {
 	pcre_malloc = (void *(*) (size_t)) mymalloc;
